/*
 * Copyright 2010 Canonical, Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of either or both of the following licenses:
 *
 * 1) the GNU Lesser General Public License version 3, as published by the
 * Free Software Foundation; and/or
 * 2) the GNU Lesser General Public License version 2.1, as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the applicable version of the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of both the GNU Lesser General Public
 * License version 3 and version 2.1 along with this program.  If not, see
 * <http://www.gnu.org/licenses/>
 *
 * Authors:
 *    Cody Russell <crussell@canonical.com>
 */
#include <Python.h>
#include "../src/gropegesturemanager.h"


/* Set up GestureManager object */
typedef struct {
    PyObject_HEAD
    PyObject *x_attr; /* Attributes dictionary */
    // XXX set up registry... or a reference to the registry? May need to tie into grope_gesture_manager_get somehow...
    //PyObject *_registry;
} GestureManagerObject;

static PyTypeObject GestureManager_type;

#define GestureManagerObject_Check(v)      (Py_TYPE(v) == &GestureManager_type)

static GestureManagerObject * newGestureManagerObject(PyObject *arg) {
    GestureManagerObject *self;
    self = PyObject_New(GestureManagerObject, &GestureManager_type);
    if (self == NULL)
        return NULL;
    self->x_attr = NULL;
    return self;
}

/* GestureManager methods */
static void GestureManager_dealloc(GestureManagerObject *self) {
    Py_XDECREF(self->x_attr);
    PyObject_Del(self);
}

static PyObject *GestureManager_register_window(GestureManagerObject *self) {
    // XXX need to pass appropirate args here
    grope_gesture_manager_register_window();
    Py_INCREF(Py_None);
    return Py_None;
}

static int GestureManager_setattr(GestureManagerObject *self,
                                  char *name, PyObject *v) {
    if (self->x_attr == NULL) {
        self->x_attr = PyDict_New();
        if (self->x_attr == NULL)
            return -1;
    }
    if (v == NULL) {
        int rv = PyDict_DelItemString(self->x_attr, name);
        if (rv < 0)
            PyErr_SetString(PyExc_AttributeError,
                "delete non-existing GestureManager attribute");
        return rv;
    }
    else
        return PyDict_SetItemString(self->x_attr, name, v);
}

static PyMethodDef GestureManager_methods[] = {
    {"register_window", (PyCFunction)GestureManager_register_window,
     METH_VARARGS, PyDoc_STR("get_thing_one() -> 1")},
    {NULL, NULL},
};

static PyObject * GestureManager_getattr(GestureManagerObject *self,
                                         char *name) {
    if (self->x_attr != NULL) {
        PyObject *v = PyDict_GetItemString(self->x_attr, name);
        if (v != NULL) {
            Py_INCREF(v);
            return v;
        }
    }
    return Py_FindMethod(GestureManager_methods, (PyObject *)self, name);
}

static PyTypeObject GestureManager_type = {
    /* The ob_type field must be initialized in the module init function
     * to be portable to Windows without using C++. */
    PyVarObject_HEAD_INIT(NULL, 0)
    "_grope.GestureManager",               /*tp_name*/
    sizeof(GestureManagerObject),           /*tp_basicsize*/
    0,                                      /*tp_itemsize*/
    /* methods */
    (destructor)GestureManager_dealloc,     /*tp_dealloc*/
    0,                                      /*tp_print*/
    (getattrfunc)GestureManager_getattr,    /*tp_getattr*/
    (setattrfunc)GestureManager_setattr,    /*tp_setattr*/
    0,                                      /*tp_compare*/
    0,                                      /*tp_repr*/
    0,                                      /*tp_as_number*/
    0,                                      /*tp_as_sequence*/
    0,                                      /*tp_as_mapping*/
    0,                                      /*tp_hash*/
    0,                                      /*tp_call*/
    0,                                      /*tp_str*/
    0,                                      /*tp_getattro*/
    0,                                      /*tp_setattro*/
    0,                                      /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT,                     /*tp_flags*/
    0,                                      /*tp_doc*/
    0,                                      /*tp_traverse*/
    0,                                      /*tp_clear*/
    0,                                      /*tp_richcompare*/
    0,                                      /*tp_weaklistoffset*/
    0,                                      /*tp_iter*/
    0,                                      /*tp_iternext*/
    0,                                      /*tp_methods*/
    0,                                      /*tp_members*/
    0,                                      /*tp_getset*/
    0,                                      /*tp_base*/
    0,                                      /*tp_dict*/
    0,                                      /*tp_descr_get*/
    0,                                      /*tp_descr_set*/
    0,                                      /*tp_dictoffset*/
    0,                                      /*tp_init*/
    0,                                      /*tp_alloc*/
    0,                                      /*tp_new*/
    0,                                      /*tp_free*/
    0,                                      /*tp_is_gc*/
};
/* --------------------------------------------------------------------- */

/* Set up the module-level class */
static PyObject * _grope_GestureManager(PyObject *self, PyObject *args) {
    GestureManagerObject *rv;

    if (!PyArg_ParseTuple(args, ":new"))
        return NULL;
    rv = newGestureManagerObject(args);
    if (rv == NULL)
        return NULL;
    return (PyObject *)rv;
}

static PyMethodDef _grope_functions[] = {
    {"GestureManager", _grope_GestureManager, METH_VARARGS,
        PyDoc_STR("GestureManager() -> new GestureManager object")},
    {NULL, NULL}
};

/* Set up a whole bunch of constants without the need for lots of boilerplate.
 */
struct _IntConstantPair {
    char *constant_name;
    int constant_value;
};

typedef struct _IntConstantPair IntConstantPair;

static IntConstantPair _constants[] = {
    {"GROPE_GESTURE_DRAG1", GROPE_GESTURE_DRAG1},
    {"GROPE_GESTURE_DRAG2", GROPE_GESTURE_DRAG2},
    {"GROPE_GESTURE_DRAG3", GROPE_GESTURE_DRAG3},
    {"GROPE_GESTURE_DRAG4", GROPE_GESTURE_DRAG4},
    {"GROPE_GESTURE_DRAG5", GROPE_GESTURE_DRAG5},
    {"GROPE_GESTURE_PINCH1", GROPE_GESTURE_PINCH1},
    {"GROPE_GESTURE_PINCH2", GROPE_GESTURE_PINCH2},
    {"GROPE_GESTURE_PINCH3", GROPE_GESTURE_PINCH3},
    {"GROPE_GESTURE_PINCH4", GROPE_GESTURE_PINCH4},
    {"GROPE_GESTURE_PINCH5", GROPE_GESTURE_PINCH5},
    {"GROPE_GESTURE_ROTATE1", GROPE_GESTURE_ROTATE1},
    {"GROPE_GESTURE_ROTATE2", GROPE_GESTURE_ROTATE2},
    {"GROPE_GESTURE_ROTATE3", GROPE_GESTURE_ROTATE3},
    {"GROPE_GESTURE_ROTATE4", GROPE_GESTURE_ROTATE4},
    {"GROPE_GESTURE_ROTATE5", GROPE_GESTURE_ROTATE5},
    {"GROPE_GESTURE_DRAG_ALL", GROPE_GESTURE_DRAG_ALL},
    {"GROPE_GESTURE_PINCH_ALL", GROPE_GESTURE_PINCH_ALL},
    {"GROPE_GESTURE_ROTATE_ALL", GROPE_GESTURE_ROTATE_ALL},
    {"GROPE_GESTURE_ALL", GROPE_GESTURE_ALL},
    {(char*)NULL, 0}
};

/* Set up the module initialization. */
PyMODINIT_FUNC
init_grope(void) {
    PyObject *module, *dict;
    module = Py_InitModule("_grope", _grope_functions);
    dict = PyModule_GetDict(module);

    /* Finalize the type object including setting type of the new type
     * object; doing it here is required for portability, too. */
    if (PyType_Ready(&GestureManager_type) < 0)
        return;

    /* Set integer constants */
    int i;
    for (i = 0; _constants[i].constant_name != 0; i++) {
        PyModule_AddIntConstant(module,
                                _constants[i].constant_name,
                                _constants[i].constant_value);
    }
}
